#!/usr/bin/env python
"""
Maildir Library
by Javantea
Sept 27, 2007 - Jan 13, 2013

Main page:
	directories on the left
	messages on the top right
	message body in the bottom right

"""
from __future__ import print_function
import os
import sys
import stat
import email
import email.utils
import email.header
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart
from email.mime.audio import MIMEAudio
from email.mime.base import MIMEBase
from email.mime.image import MIMEImage
from email.mime.application import MIMEApplication
import time
import io
import re
import configparser
import zlib
import json
import socket
import string
import hashlib
try:
	from html import unescape as entity_unescape
except ImportError:
	entity_unescape = None
#end try
if entity_unescape == None:
	try:
		from HTMLParser import HTMLParser
		entity_unescape = HTMLParser().unescape
	except ImportError:
		print("Warning: You lack a valid html entity parser. You may get ugly markdown.", file=sys.stderr)
		def entity_unescape(s):
			"""
			Deal with the most common issue in urls, ampersand (&).
			"""
			# TODO: implement a fully functional html entity decoder.
			return s.replace('&amp;', '&')
		#end def entity_unescape(s)
	#end try
#end if

data_dir = os.path.split(__file__)[0]

try:
	# Optional GPG encryption, decryption, sign and verify
	import bikeim.gpg3 as gpg3
except ImportError:
	gpg3 = None
#end try
try:
	import dateutil.parser
except ImportError:
	dateutil = None
#end try

PY2 = False
if sys.hexversion >= 0x3000000:
	BytesIO = io.BytesIO
	baseUnicode = str
else:
	PY2 = True
	class FileNotFoundError(OSError):
		pass
	class NotADirectoryError(OSError):
		pass
	class FileExistsError(OSError):
		pass
	BytesIO = io.StringIO
	baseUnicode = unicode
#end if

try:
	from lxml.html.clean import Cleaner
	from lxml import html
	def clean_html(s):
		if s == '' or s == None: return ''
		xclean = Cleaner(style=True, add_nofollow=True, kill_tags=['img'], remove_tags=['table','tr','td','th','thead','tbody','div','span','strong','b','em','i','center','font','p','h1','h2','h3','h4','h5','h6','hr','body','div'])
		ch = xclean.clean_html(s)
		ch = brFix(ch)
		ch = re.sub(r'</div>$', '', re.sub(r'^<div>', '', ch))
		return re.sub(r'</p>$', '', re.sub(r'^<p>', '', ch))
	#end def clean_html(s)
	def text_html(s):
		if s == '' or s == None: return ''
		tree = html.fromstring(s)
		yclean = Cleaner()
		lclean_html = yclean.clean_html
		tree = lclean_html(tree)
		text = tree.text_content()
		return text
	#end def text_html(s)
except ImportError:
	# If we don't have lxml, we use our own =D
	def clean_html(s):
		"""
		Not very good, but practically good enough.
		"""
		return encodeltgt(s)
	#end def clean_html(s)
	def text_html(s):
		"""
		Not very good, but practically good enough.
		"""
		return encodeltgt(s)
	#end def text_html(s)
#end try

anchor_re = re.compile(r"<a\s.*?\s?href=['\"]([^'\"]*)[^>]*>(.*?)</a>", re.MULTILINE|re.DOTALL|re.IGNORECASE)
whitespace_re = re.compile(r"(\s)\s*", re.MULTILINE)
newline_re = re.compile(r'\n[\n\t\xa0 ]*', flags=re.MULTILINE)
email_re = re.compile('("[^<]*")? ?<([^>,;]*)>|([^<,;]*)? ?<([^>,;]*)>|([^,;]*)', re.MULTILINE|re.DOTALL)
maildir_dir_re = re.compile(r'^\.(.*?)\.directory/')
# RFC5322 3.2.3. Atom atext
display_name_re = re.compile('^[' + re.escape(string.ascii_lowercase + string.ascii_uppercase + string.digits + '!#$%&\'*+-/=?^=`{|}~ ') + ']*$')
br_re = re.compile(r"<br[^>]*>", re.MULTILINE|re.DOTALL|re.IGNORECASE)

def deWhitespace(s):
	"""
	Take any whitespace at the beginning and end and remove.
	Take any dual whitespace in the middle and make it one.
	"""
	s = s.strip()
	return whitespace_re.sub(r' ', s)
#end def deWhitespace(s)

def subNewline(s, blank=False):
	"""
	Replace "\n" + whitespace with "\n" or ""
	"""
	substitution = '\n'
	if blank: substitution = ''
	return newline_re.sub(substitution, s)
#end def subNewline(s)

def markdownEscape(s):
	"""
	Currently there is no character I know I need to escape in markdown.
	"""
	return s
#end def markdownEscape(s)

def brFix(s):
	"""
	Replace <br> and all similar with newline.
	TODO: Too general, also matches <brain>
	"""
	return br_re.sub('\n', s)
#end def brFix(s)

def anchorToMarkdown(s):
	"""
	Similar to anchor_re.sub(r'[\3](\1)', s)) but with whitespace removal.
	"""
	output = s
	while True:
		res = anchor_re.search(output)
		if not res: break
		start, end = res.span()
		link, text = res.groups()
		link = entity_unescape(link)
		output = output[:start] + '[' + markdownEscape(deWhitespace(text)) + '](' + link + ')' + output[end:]
	#loop
	output = entity_unescape(output)
	return output
#end def anchorToMarkdown(s)

def anchorToFootnote(s):
	"""
	Similar to anchorToMarkdown but using footnotes like:
	<a href="https://example.com">test</a>
	test[1]
	[1] https://example.com
	"""
	output = s
	footnotes = ''
	footnote_i = 1
	while True:
		res = anchor_re.search(output)
		if not res: break
		start, end = res.span()
		link, text = res.groups()
		footnote = ('[{0}]'.format(footnote_i))
		output = output[:start] + markdownEscape(deWhitespace(text)) + footnote + output[end:]
		footnotes += footnote + ' ' + link + '\n'
		footnote_i += 1
	#loop
	if output[-1:] != '\n': output += '\n'
	output += footnotes
	output = entity_unescape(output)
	#print('output', repr(output))
	return output
#end def anchorToFootnote(s)

# TODO: Put this into a config variable.
if 'HOME' in os.environ:
	home = os.environ['HOME']
else:
	home = os.path.expanduser('~')
#end if
hidden_dir = os.path.join(home, '.mail', '')

def getConfig():
	"""
	Wrapper for configparser so that everyone can get uniform and extensible access.
	Format TBD
	"""
	filename = os.path.join(hidden_dir, 'mail1rc')
	config = configparser.ConfigParser()
	if not os.path.exists(filename): return config
	config.read_file(open(filename, 'r'))
	return config
#end def getConfig()

def saveConfig(config):
	"""
	Wrapper for configparser so that everyone can get uniform and extensible access.
	Format TBD
	"""
	config.write(open(os.path.join(hidden_dir, 'mail1rc'), 'w'))
	return config
#end def saveConfig()

KNOWN_MULTIPART = ['multipart/related', 'multipart/mixed', 
	'multipart/alternative', 'multipart/signed', 'multipart/encrypted', 
	'multipart/report']

def toUnicode(s):
	"""
	Convert anything to unicode.
	"""
	if isinstance(s, baseUnicode):
		return s
	elif isinstance(s, bytes):
		# TODO: Deal with bad stuff better because we can.
		return s.decode('utf-8', 'replace')
	#end if
	return baseUnicode(s)
#end def toUnicode(s)

def sanitizeCli(s):
	"""
	CLI needs no encoding. HTML does.
	"""
	if s == None: return ''
	# FIXME: why does this need space?
	return deWhitespace(s.replace('\n', ' '))
#end def sanitizeCli(s)

def decodeHeaderCli(h, toSanitize=True, debug=False):
	"""
	We don't need to sanitize in Gtk. Only in HTML.
	"""
	if h == None:
		if toSanitize: return sanitizeCli(h)
		return ''
	#end if
	if isinstance(h, email.header.Header):
		h = baseUnicode(h)
	#end if
	if h.strip()[:2] == '=?' or '=?' in h:
		output = decodeQE(h, debug)
	else:
		# Seems legit... How could this be a bad choice? Oh yeah.. 
		# TODO: Ensure that we don't get hacked putting a decoder and a sanitizer together. Shit.
		if not isinstance(h, baseUnicode):
			output = guessEncoding(h)
		else:
			output = h
		#end if
	#end if
	if toSanitize:
		return sanitizeCli(output)
	#end if
	return output
#end def decodeHeaderCli(s, toSanitize=False, debug=False)

def decodeDateHeader(date_header):
	"""
	Homogenize dates.
	Input:
	'Sun, 30 Apr 2017 10:05:47 -0700'
	Output:
	'Sun, 30 Apr 2017 10:05:47 -0700'
	FIXME: This is very broken.
	"""
	if dateutil:
		try:
			time_split = dateutil.parser.parse(date_header).timetuple()
		except ValueError:
			return date_header
		#end try
		if time_split.tm_gmtoff:
			local_time = time.gmtime(time.mktime(time_split) + time_split.tm_gmtoff)
		else:
			local_time = time.gmtime(time.mktime(time_split))
		#end if
	else:
		time_split = time.strptime(date_header, '%a, %d %b %Y %H:%M:%S %z')
		local_time = time.gmtime(time.mktime(time_split) + time_split.tm_gmtoff)
	#end if
	return time.strftime('%a, %d %b %Y %H:%M:%S', local_time)
#end def decodeDateHeader(date_header)

def readLine(fp, maxlen=None):
	"""
	This is a more expensive average readline, but it avoids the worst case 
	where the file is simply a 4GB binary file with no newlines.
	This is very similar to fgets except that it doesn't terminate at null.
	The only penalty to making maxlen megabytes is the repeated tell/read/seek 
	which is also a nontrivial problem when maxlen is small.
	readLine(fp) is similar to fp.readline()
	"""
	if not fp.seekable():
		raise ValueError("File is unable to seek, we can't readlines.")
	#end if
	data = fp.read(maxlen)
	if b'\n' not in data:
		return data
	#end if
	eol = data.index(b'\n') + 1
	ret = data[:eol]
	fp.seek(eol-len(data), io.SEEK_CUR)
	return ret
#end def readLine(fp, maxlen=None)

def unbin(s):
	"""
	TODO: Understand this better. For now it's a workaround for semi-ascii 
	headers.
	"""
	return s.decode('ascii', 'surrogateescape').strip()
#end def unbin(s)

def smartEmailParseFile(fp, filename=None):
	"""
	Create a message from a file pointer without recursion. Easy enough.
	"""
	m = email.message.Message()
	if fp.closed:
		if filename == None:
			return m
		#end if
		fp = open(filename, 'rb')
	#end if
	fp.seek(0, io.SEEK_SET)
	headerToAdd = None
	eof = False
	while not eof:
		# Always increasing, never an endless while loop.
		# Any line longer than 65536 is considered two lines.
		data = readLine(fp, 0x10000)
		if data == b'':
			eof = True
			break
		#end if
		if data[-1:] != b'\n':
			# Long line means that we should stop parsing for now.
			fp.seek(-len(data), io.SEEK_CUR)
			break
		#end if
		# Strip off the newline
		data = data[:-1]
		if data == b'' or data == b'\r':
			print('end of headers?', data)
			break
		#end if
		if (data[:1] == b'\t' or data[:1] == b' ') and headerToAdd != None:
			# Additional data for the previous header.
			headerToAdd += b'\n' + data
		elif b':' in data:
			# We can decode with ascii because email headers are not binary or are they? Who knows?
			if headerToAdd != None:
				# Add the previous one.
				lhs, rhs = headerToAdd.split(b':', 1)
				m[unbin(lhs)] = unbin(rhs)
			#end if
			headerToAdd = data
		else:
			print("Invalid header?", data)
		#end if
	#loop
	# Add the previous one.
	if headerToAdd != None:
		lhs, rhs = headerToAdd.split(b':', 1)
		m[unbin(lhs)] = unbin(rhs)
	#end if
	# Set the content type so that we can communicate the data to the user. We 
	# could do a binary with a name so that they could download it and reopen 
	# it. But that just begs for recursion again.
	if 'content-type' in m:
		m.replace_header('content-type', 'text/plain')
	else:
		m['Content-Type'] = 'text/plain'
	#end if
	pl = "This crashed Python's parser, so we are using smartEmailParse(). "
	pl += "Here is the data still encoded:\n\n"
	# FIXME: Definitely don't want ascii/replace here. Better than nothing.
	#raise Exception("smartEmailParseFile")
	#pl += repr(fp.read()) #.decode('ascii', 'replace')
	pl += fp.read().decode('utf-8', 'surrogateescape')
	m.set_payload(pl)
	
	return m
#end def smartEmailParseFile(fp, filename=None)

def smartEmailParse(s):
	"""
	Create a message from a file pointer without recursion. Easy enough.
	TODO: Test this fully.
	"""
	fp = BytesIO(s)
	return smartEmailParseFile(fp)
#end def smartEmailParse(s)

def message_from_binary_file(fp, filename=None, *args, **kws):
	"""
	Simple wrapper for email.message_from_binary_file that supports Python 2 and 3.
	"""
	try:
		if sys.hexversion >= 0x3020000:
			# This function doesn't exist in Python2.7, Python3.0, or Python3.1.
			return email.message_from_binary_file(fp, *args, **kws)
		else:
			# FIXME: 3.0 and 3.1!
			return email.message_from_file(fp, *args, **kws)
		#end if
	except RuntimeError as rte:
		if rte.args and (len(rte.args) == 1) and (rte.args[0] == 'maximum recursion depth exceeded'):
			print("We found a recursion error, hooray. Run it again without the except to get details.")
			if filename: print(filename)
			# Many levels of recursion in the message, so it won't work.
			return smartEmailParseFile(fp, filename)
		#end if
		print(rte)
		print(dir(rte))
		print(repr(rte.args))
	#except OSError as ose:
	#	# An odd handler in case we want to debug.
	#	raise ose
	#end try
	return smartEmailParseFile(fp, filename)
#end def message_from_binary_file(fp, filename=None, *args, **kws)

def message_from_bytes(s, *args, **kws):
	"""
	Simple wrapper for email.message_from_bytes that supports Python 2 and 3.
	This is a bad workaround for email.message_from_file because it means a 
	very large file gets loaded into memory. This might be fixed by mmap or 
	something clever. The reason we have to do message_from_bytes is because 
	we don't know what encoding a person is using when they send us an e-mail 
	until after we've loaded each part.
	FIXME: too much data problem.
	To use this function:
	f = open(fname, 'rb')
	fdata = f.read()
	# This takes too long, need to use index or something...
	msg = mail1.message_from_bytes(fdata)
	"""
	try:
		if sys.hexversion >= 0x3020000:
			# This function doesn't exist in Python2.7.
			return email.message_from_bytes(s, *args, **kws)
		else:
			# FIXME: 3.0 and 3.1!
			return email.message_from_string(s, *args, **kws)
		#end if
	except RuntimeError as rte:
		if rte.args and (len(rte.args) == 1) and (rte.args[0] == 'maximum recursion depth exceeded'):
			print("We found a recursion error, hooray. Run it again without the except to get details.")
			# Many levels of recursion in the message, so it won't work.
			return smartEmailParse(s)
		#end if
		print(rte)
		print(dir(rte))
		print(repr(rte.args))
	#except OSError as ose:
	#	# An odd handler in case we want to debug.
	#	raise ose
	#end try
	return smartEmailParse(s)
#end def message_from_bytes(s, *args, **kws)

def encodeltgt(a):
	"""
	AAAAnnd done =)
	"""
	return a.replace('>', '&gt;').replace('<', '&lt;')
#end def encodeltgt(a)

spam_color = ('#00ff00', '#33ff00', '#66ff00', '#99ff00', '#ccff00', 
	'#ffff00', '#ffcc00', '#ff9900', '#ff6600', '#ff3300', '#ff0000')

def decodeQE(h, debug=False):
	"""
	Convert a =?utf-8?Q?Team=20Uber=20Seattle?= style header to a string.
	Output should be: 'Team Uber Seattle'
	Handles all possible exceptions.
	FIXME: Does not work when you give it intentionally invalid data.
	It may be impossible to fix intentionally invalid data.
	"""
	output = ''
	pos = 0
	while '=?' in h[pos:]:
		start = h.index('=?', pos)
		charset_end = h.index('?', start+2)
		if '?=' in h[charset_end+3:]:
			end = h.index('?=', charset_end+3)+2
		else:
			end = len(h)
		#end if
		pre = h[pos:start]
		#print('pre', repr(pre))
		output += subNewline(pre, True)
		#if debug:
		#	return h + '[%i %i %s]' % (start, end, pre)
		j = h[start:end]
		#h_decoded = email.header.decode_header(j.strip().replace('\n', ''))
		h_decoded = email.header.decode_header(subNewline(j.strip(), True))
		# Nothing found
		if h_decoded[0][1] == None:
			return h
		#end if
		h_uni = guessEncoding(h_decoded[0][0], h_decoded[0][1])
		if h_uni.success:
			# Adding the encoding breaks threading. See #nottor
			output += h_uni #+ ' ' + ('(%s)' % h_decoded[0][1])
		else:
			if debug:
				# Return repr so we can debug the thing in the program/output.
				output += 'Error({0!r}, {1!r})'.format(h, '') # str(ude)
			else:
				# Use the string we guess using utf-8 or cp1252
				output += h_uni
			#end if
		#end if
		pos = end
	#loop
	#print('unparsed', repr(h[pos:]))
	
	return output + h[pos:]
#end def decodeQE(h, debug=False)

class successStr(baseUnicode):
	"""
	A subclass of str that acts like str, but has a hidden attribute, 
	success. You can guess the name of the attribute from the name of the 
	class or by doing a dir.
	"""
	def __init__(self, value, success=False):
		#print(repr(value))
		#baseUnicode.__init__(self, value)
		self.success = success
	#end def __init__(value, success=False)
	def __new__(cls, value, success=False):
		#print('new', repr(value))
		return baseUnicode.__new__(cls, value)
	#end def __new__(value, success=False)
#end class successStr(baseUnicode)

def guessEncoding(value, charset=None):
	"""
	Converts a bytes value into a string with the most likely charset.
	Also optionally returns whether the encoding succeeded.
	It's too bad that e-mail software doesn't know which fragging encoding it's 
	sending... or this wouldn't be necessary.
	"""
	if value == None:
		return ''
	#end if
	if charset == None:
		# Default to the most sane multi-byte encoding...
		charset = 'utf-8'
	#end if
	# We're going to use the HTML5 spec here...
	# https://encoding.spec.whatwg.org/#names-and-labels
	if charset.lower().replace('-','') == 'iso88591':
		charset = 'cp1252'
	#end if
	success = True
	try:
		ret = value.decode(charset)
	except LookupError:
		success = False
	except UnicodeDecodeError:
		success = False
	#end try
	if not success:
		if detectWindows1252(value):
			# Spammers use this so often that it's worth doing this.
			ret = value.decode('cp1252', 'replace')
		else:
			ret = value.decode('utf-8', 'replace')
		#end if
	#end if
	# Report whether we succeeded.
	return successStr(ret, success)
#end def guessEncoding(value, charset=None)

def sanitize(data):
	"""
	Does not actually sanitize HTML or bytes. Be careful using this in a web context!
	"""
	if data == None: return '**None.**'
	if data == '': return '*Empty.*'
	return data
#end def sanitize(data)

def readDir(filename=None, recurse=False, tabs=0):
	"""
	Take a directory listing (optionally recursive) and sort it into 
	directories, files, hidden files, and hidden directories.
	"""
	output = {'dirs':[], 'files':[], 'hidden':[], 'hidden dirs':[]}
	try:
		files = os.listdir(filename)
	except FileNotFoundError:
		print("Error: Missing directory", filename)
		return output
	#end try
	files.sort()

	for fi in files:
		s = os.stat(os.path.join(filename, fi))
		if stat.S_ISDIR(s[stat.ST_MODE]):
			where = 'dirs'
			if fi[0] == '.':
				where = 'hidden dirs'
			#end if
			d = ('\t' * tabs) + fi + os.sep
			if recurse:
				e = readDir(os.path.join(filename, fi), True, tabs + 1)
				output[where].append((d, e))
			else:
				output[where].append(d)
			#end if
		elif stat.S_ISREG(s[stat.ST_MODE]):
			where = 'files'
			if fi[0] == '.':
				where = 'hidden'
			#end if
			output[where].append(('\t' * tabs) + fi)
		else:
			# Probably a symlink, *shrug*
			output['files'].append(('\t' * tabs) + fi) # + '?')
		#end if
	#next fi
	return output
#end def readDir(filename=None, recurse=False, tabs=0)

def filterDir(filter_filename, path=None, currentDir=''):
	"""
	Filter every file in a directory.
	"""
	if path == None:
		#path = '/home/jvoss/Mail'
		path = os.path.join(os.path.expanduser('~'), 'Mail')
	#end if
	messages = readDir(os.path.join(path, currentDir, 'cur'))
	files = messages['files']
	return filterList(filter_filename, files, path, currentDir)
#end def filterDir(filter_filename, path = '/home/$USER/Mail', currentDir = '')

def searchBody(search_term, message, case_sensitive=False):
	"""
	search_term is a string you want to find.
	message is a filename.
	"""
	# TODO: Any other readable 
	readable = ['text/plain', 'text/html']
	if not case_sensitive:
		search_term = search_term.lower()
	#end if
	try:
		f = open(message, 'rb')
	except FileNotFoundError:
		return False
	#end if
	msg = message_from_binary_file(f, message)
	for part in msg.walk():
		ctype = part.get_content_type()
		if ctype in readable:
			pdata = part.get_payload(decode=True)
			# Copied from mail1glade2.py:
			ct = part.get('Content-Type')
			charset = None
			if ct and 'charset=' in ct:
				charset = ct[ct.index('charset=')+8:]
			#end if
			# Why not just errors='replace'? Encoding is more 
			# valuable than that.
			pdata = guessEncoding(pdata, charset)
			if not pdata.success:
				# warning.
				#pdata = "Error: decoding failed, it's spam =)\n\n" + pdata
				pass
			#end if
			if not case_sensitive:
				pdata = pdata.lower()
			#end if
			if search_term in pdata:
				return True
			#end if
		#end if
	#next part
	return False
#end def searchBody(search_term, message, case_sensitive=False)

def filterList(filter_filename, files, path=None, currentDir=''):
	"""
	Filter a list of files using a filter.
	filter_filename is the filename of the filter in hidden_dir ~/.mail/.
	files is a list of relative filenames.
	path is the first part of the path, optional ~/Mail if left None.
	currentDir is the second part of the path optional '' if left ''.
	FIXME: This code is poorly written, but works well enough.
	"""
	if path == None:
		#path = '/home/jvoss/Mail'
		path = os.path.join(home, 'Mail')
	#end if
	"""
	filters = {'nouveau':{'List-Id':'nouveau.lists.freedesktop.org', 'dir':'altsci/nouveau'}, 
		}
	"""
	if ('..' + os.sep) in filter_filename:
		print("Error: Unacceptable filename for filter.")
		return "Error: Unacceptable filename for filter.", []
	#end if
	filters = {}
	filter_list = open(os.path.join(hidden_dir, 'filter', filter_filename), 'r').read().split('\n')
	for filter_line in filter_list:
		if filter_line.strip() == '': continue
		x = json.loads(filter_line)
		if len(x) != 4:
			print("Warning: {0}".format(filter_filename), repr(filter_line))
			continue
		#end if
		name, field, value, loc = x
		if name in filters:
			print("Warning: {0}".format(filter_filename), repr(name), 'in list twice')
			continue
		#end if
		filters[name] = {field:value, 'dir':loc}
	#next filter_line
	output = ''
	moves = []
	# messages on the top right
	#show = show[page*pagesize:(page+1)*pagesize]
	for message in files:
		fname = os.path.join(path, currentDir, 'cur', message)
		f = open(fname, 'rb')
		msg = message_from_binary_file(f, fname)
		subject = str(msg['subject'])
		if subject and (subject.strip()[:2] == '=?'):
			try:
				subject = sanitize(email.header.decode_header(subject.strip().replace('\n', ''))[0][0])
			except Exception as e:
				subject = str(e) + '*' + sanitize(subject)
			#end try
		else:
			subject = sanitize(subject)
		#end if
		mfrom = decodeHeaderCli(msg['from'])
		mto = decodeHeaderCli(msg['to'])
		mcc = decodeHeaderCli(msg['cc'])
		mlistid = decodeHeaderCli(msg['List-Id'])
		#date = sanitize(str(msg['date']))
		sfrom = decodeHeaderCli(msg['from'])
		sto = decodeHeaderCli(msg['to'])
		scc = decodeHeaderCli(msg['cc'])
		slistid = decodeHeaderCli(msg['List-Id'])
		#sdate = str(msg['date'])
		ssubject = decodeHeaderCli(msg['subject'])
		match_name = None
		match_dir = None
		for fname, filterd in filters.items():
			if 'List-Id' in filterd:
				if slistid and filterd['List-Id'].lower() in slistid.lower():
					match_name = fname
					match_dir = filterd['dir']
				#end if
			elif 'To' in filterd:
				if sto and filterd['To'].lower() in sto.lower():
					match_name = fname
					match_dir = filterd['dir']
				#end if
			elif 'From' in filterd:
				if sfrom and filterd['From'].lower() in sfrom.lower():
					match_name = fname
					match_dir = filterd['dir']
				#end if
			elif 'Cc' in filterd:
				if scc and filterd['Cc'].lower() in scc.lower():
					match_name = fname
					match_dir = filterd['dir']
				#end if
			elif 'Subject' in filterd:
				if ssubject and filterd['Subject'].lower() in ssubject.lower():
					match_name = fname
					match_dir = filterd['dir']
				#end if
			else:
				output += 'Error: broken filter,' + sanitize(repr(filterd)) + '\n'
			#end if
			if match_name:
				break
			#end if
		#next filter
		if match_name:
			actual_match_dir = ''
			match_dir_list = match_dir.split(os.sep)
			if len(match_dir_list) > 0 and match_dir_list[-1] == '':
				match_dir_list.pop()
			#end if
			for directory in match_dir_list[:-1]:
				actual_match_dir += '.' + directory + '.directory' + os.sep
			#next directory
			actual_match_dir += match_dir_list[-1]
			output += 'Matched ' + match_name + ' ' + actual_match_dir + '\n'
			# Should never happen unless it gets screwy
			if os.path.exists(os.path.join(path, actual_match_dir, 'cur', message)):
				output += 'Error: File exists in destination, not moving: ' + os.path.join(path, actual_match_dir, 'cur', message) + '\n'
				continue
			#end if
			# Rare but could happen if the filter was made before the dir or no cur
			if not os.path.exists(os.path.join(path, actual_match_dir, 'cur')):
				#output += 'Error: Dest dir missing, not moving: ' + os.path.join(path, actual_match_dir, 'cur') + '\n'
				output += 'Warning: Dest dir missing, creating: ' + os.path.join(path, actual_match_dir, 'cur') + '\n'
				try:
					os.makedirs(os.path.join(path, actual_match_dir, 'cur'))
				except (FileExistsError, PermissionError) as e:
					output += "Error on makedirs: " + str(e)
					return output, moves
				#end try
			#end if
			os.rename(os.path.join(path, currentDir, 'cur', message), os.path.join(path, actual_match_dir, 'cur', message))
			moves.append((message, os.path.join(path, actual_match_dir, 'cur')))
		else:
			#output += '<tr><td><div class="s"><a href="mail1mod.py?dir=' + currentDir + '&msg=' + message + '" title="' + subject + '" target="msgf">' + spam_icon + ssubject + '</a></div></td><td><div class="f" title="' + mfrom + '">' + smfrom + '</div></td><td><div class="d" title="' + date + '">' + date + '</div></td></tr>\n'
			output += 'From: ' + mfrom + '\n' + 'To: ' + mto + '\n' + 'Cc: ' + mcc + '\n' + 'List-Id: ' + mlistid + '\n'
		#end if
	#next message
	return output, moves
#end def filterList(files, path = '/home/$USER/Mail', currentDir = '')

def moveAll(srcDir, destDir):
	"""
	Move all files from source to destination.
	Returns a list of filenames relative to destination.
	"""
	messages = readDir(srcDir)
	files = messages['files']
	moved = []
	for message in files:
		srcPath = os.path.join(srcDir, message)
		destPath = os.path.join(destDir, message)
		if os.path.exists(destPath):
			print('Error: dest exists:', destPath)
			continue
		#end if
		print('renamed', srcPath, destPath)
		os.rename(srcPath, destPath)
		# chmod 644 file, user can read and write, group can read, other can read.
		try:
			os.chmod(destPath, stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP | stat.S_IROTH)
		except (OSError, IOError) as e:
			print("Warning:", e)
		#end try
		moved.append(message)
	#next message
	return moved
#end def moveAll(srcDir, destDir)

def nameSanitize(name):
	"""
	Name --> Name
	Javantea Test --> "Javantea Test"
	Whatever" --> "Whatever\""
	"""
	# RFC5322 3.2.3. Atom atext
	if not display_name_re.match(name):
		name = name.replace('\\', '\\\\').replace('"', '\\"')
		name = '"' + name + '"'
	#end if
	return name
#end def nameSanitize(name)

def canonTo(to):
	"""
	Ensure that to is an array
	Ensure sto is either None or a valid string.
	Valid strings look like 'a <a@example.com>, b <b@example.com>'
	"""
	if PY2 and isinstance(to, str):
		to = to.decode('utf-8')
	#end if
	if isinstance(to, baseUnicode):
		pos = 0
		#print('addr', addr.__class__, addr)
		addr_match = email_re.match(to)
		to_list = []
		while addr_match:
			name, addr, name2, addr2, addr3 = addr_match.groups()
			if not name and not addr and not name2 and not addr2 and not addr3: break
			pos += addr_match.span()[1] + 1
			if not addr:
				addr = addr2
				name = name2
			#end if
			if not addr:
				addr = addr3
				name = ''
			#end if
			addr = addr.replace('\n', '').strip()
			if not name: name = addr
			name = name.strip().strip('"').strip()
			if addr == '':
				print('ERROR: Invalid receipient', to)
				if pos >= (len(to) - 1):
					break
				#end if
				continue
			#end if
			if name == addr:
				name = ''
			#end if
			if name == '':
				recipient_str = addr
			else:
				recipient_str = '{0} <{1}>'.format(nameSanitize(name), addr)
			#end if
			to_list.append(recipient_str)
			#print('=>', name, addr)
			if pos >= len(to): break
			addr_match = email_re.match(to[pos:])
		#loop
		to = to_list
		sto = ', '.join(to)
	elif isinstance(to, list):
		# TODO: do these values need to be sanitized?
		sto = ', '.join(to)
	elif isinstance(to, tuple):
		# TODO: do these values need to be sanitized?
		sto = ', '.join(to)
		to = list(to)
	elif to == '' or to == None:
		to = []
		sto = None
	else:
		raise ValueError("Argument 'to' is invalid")
	#end if
	return to, sto
#end def canonTo(to)

def getIndex(currentDir, page):
	"""
	Setup the index in the hidden dir.
	"""
	index = {}
	if not os.path.exists(hidden_dir):
		return {}
	#end if
	# The odds of a crc32 colliding unintentionally is approximately the 
	# same as the odds of the first 32-bits of sha1 colliding. Collision of 
	# crc32 is just much easier.
	currentDir = currentDir.rstrip(os.sep)
	index_name = currentDir.replace(os.sep, '_') + '.' + str(page) + '.' + ('{0:08x}'.format(zlib.crc32(currentDir.encode('utf-8'))))
	index_path = os.path.join(hidden_dir, index_name + '.idx.json')
	if os.path.exists(index_path):
		try:
			index = json.load(open(index_path, 'r'))
		except ValueError as ve:
			print("Error: json.load failed", ve)
			# Any error can be overcome by accepting it
			index = {}
		#end try
		#index_lines = open(index_path, 'r').read().split('\n')
		#for line in index_lines:
		#	x = reverserepr.reverseRepr(line)
		#	# Invalid lines
		#	if x == None or len(x) != 2: continue
		#	index[x[0]] = x[1]
		#next line
	#end if
	return index
#end def getIndex(currentDir, page)

def saveIndex(currentDir, page, index):
	# Setup the index in the hidden dir.
	if not os.path.exists(hidden_dir):
		os.mkdir(hidden_dir)
		if not os.path.exists(hidden_dir):
			# Survive inability to create ~/.mail/.
			return False
		#end if
	#end if
	# The odds of a crc32 colliding unintentionally is approximately the 
	# same as the odds of the first 32-bits of sha1 colliding. Collision of 
	# crc32 is just much easier.
	index_name = currentDir.rstrip(os.sep).replace(os.sep, '_') + '.' + str(page) + '.' + ('{0:08x}'.format(zlib.crc32(currentDir.rstrip(os.sep).encode('utf-8'))))
	index_path = os.path.join(hidden_dir, index_name + '.idx.json')
	index_file = open(index_path, 'w')
	json.dump(index, index_file)
	"""
	for x in index:
		line = repr((x, index[x]))
		index_file.write(line + '\n')
	#next line
	"""
	index_file.close()
	
	return True
#end def saveIndex(currentDir, page, index)

def decryptPayload(payload, keyid):
	"""
	Code directly from cryptoml-a6bcb044f830f5bf33520a9ba3721bb7fe8f328b
	Decrypts a -----BEGIN PGP MESSAGE----- blob.
	Checks for signature using checkSig or gpg3.GPG_decrypt.
	TODO: Should we return the result instead of sig and good_msg?
	"""
	# It should be obvious that this is necessary but sometimes.
	if gpg3 == None:
		print("Warning: GPG3 module is missing and there's data to decrypt.")
		return None
	#end if
	if not isinstance(payload, baseUnicode):
		payload = payload.decode('utf-8')
	#end if
	if not '-----BEGIN PGP MESSAGE-----' in payload:
		return None
	#end if
	try:
		result = gpg3.GPG_decrypt(payload)
	except (OSError, IOError) as ose:
		print('Warning:', ose, file=sys.stderr)
		return None
	#end try
	if result == None or len(result) != 2:
		print("Error: GPG_decrypt result is invalid", result)
		return None
	#end if
	r, plaintext = result
	plaintext = plaintext.decode('utf-8', 'surrogateescape')
	good_msg = ''
	summary = None
	if not hasattr(r, 'signatures') or len(r.signatures) == 0:
		good_sig = False
	else:
		summary = r.signatures[0].summary
		good_sig = summary & gpg3.gpg.constants.sigsum.VALID
		message = ''
		if not good_sig:
			message = ' ' + gpg3.summary_to_string(summary)
		#end if
		#print('summary 2', summary, message)
		fpr = r.signatures[0].fpr
		created = r.signatures[0].timestamp
		if r.signatures[0].key:
			name = str(r.signatures[0].key)
		else:
			keys = gpg3.getKeylist(fpr)
			if len(keys) == 1:
				name = ', '.join(['{0} <{1}>'.format(uid.name, uid.email) for uid in keys[0].uids])
			elif len(keys) > 1:
				name = 'Multiple keys: '
				for key_i in keys:
					name += ', '.join(['{0} <{1}>'.format(uid.name, uid.email) for uid in key_i.uids]) + '; '
				#next key_i
			else:
				name = 'Unknown ' + fpr
			#end if
		#end if
		sig = ('mail1: Signature made {0}\nPrimary key fingerprint: ' + fpr).format(time.strftime('%a %d %b %Y %T %Z', time.localtime(float(created))))
		good_msg = 'Good signature from {0}'.format(name) + message
	#end if
	fpr = None
	if not good_msg:
		result2 = checkSig(plaintext)
		if result2 == None or len(result2) != 5:
			# Encrypted but not signed.
			sig = "Warning: this message was not signed."
			good_msg = '' #+ str(dir(r)) + " " + repr(r.signatures)
			return (plaintext, good_sig, sig, good_msg, summary)
		#end if
		(good_sig, fpr, created, name, plaintext) = result2
		if not good_sig:
			print("Error: bad signature", repr(result))
			return None
		#end if
		# TODO: A proper message should also know trustworthyness.
		good_msg = 'Good signature from {0}'.format(name)
		sig = ('mail1: Signature made {0}\nPrimary key fingerprint: ' + fpr).format(time.strftime('%a %d %b %Y %T %Z', time.localtime(float(created))))
	else:
		# Grab fpr from sig
		fpr = r.signatures[0].fpr
	#end if
	
	if fpr == None:
		# This should never happen but...
		# FIXME: it did.
		print("Did not get a fingerprint, we can't find the user.")
		return None
	#end if
	return (plaintext, good_sig, sig, good_msg, summary)
#end def decryptPayload(payload)

def checkSig(plaintext):
	"""
	Calls gpg3.GPG_verify on signed plaintext in RFC 3156 mode. *crosses fingers*
	Copied from cryptoml verbatim. TODO: don't copy.
	"""
	# split on boundary
	if 'boundary=' not in plaintext:
		# We don't have a multipart/signed
		#print('Warning: Message has no boundary=', plaintext)
		return None
	#end if
	if 'application/pgp-signature' not in plaintext:
		# Not signed.
		return None
	#end if
	lines = plaintext.split('\n')
	content_type = lines[0]
	i = 0
	while 'boundary=' not in content_type:
		i += 1
		if i >= len(lines):
			break
		#end if
		content_type += lines[i]
	#loop
	boundary = content_type[content_type.index('boundary=')+9:]
	if ';' in boundary:
		boundary = boundary[:boundary.index(';')]
	#end if
	if ' ' in boundary:
		boundary = boundary[:boundary.index(' ')]
	#end if
	if '"' in boundary:
		boundary = boundary.replace('"','')
	#end if
	sig = None
	data = None
	# FIXME: Kluge for end of line
	sig_parts = plaintext.split('\n--' + boundary + '\n')
	sig_parts2 = plaintext.split('\r\n--' + boundary + '\r\n')
	if len(sig_parts2) > len(sig_parts):
		sig_parts = sig_parts2
	#end if
	for sig_part in sig_parts:
		lines = sig_part.split('\n')
		# FIXME: This is janky, I need a proper function here.
		if lines[0].startswith('Content-Type:') and 'multipart/signed' in lines[0]:
			continue
		#end if
		#if lines[0].startswith('Content-Type:') and 'multipart/mixed' in lines[0]:
		#	continue
		#end if
		if lines[0].startswith('Content-Type:') and 'application/pgp-signature' in lines[0]:
			sig = sig_part
			if '\n\n' in sig:
				sig = sig[sig.index('\n\n')+2:]
			elif '\r\n\r\n' in sig:
				sig = sig[sig.index('\r\n\r\n')+4:]
			#end if
			continue
		#end if
		if data != None:
			print('Warning: multiple datas', repr(plaintext))
			#print(repr(data), repr(sig_part))
			return None
		#end if
		data = sig_part
	#next sig_part
	if sig == None or data == None:
		print("Warning: No data or signature found", repr(plaintext))
		return None
	#end if
	
	# This function calls our C code.
	#print('GPG_verify2', repr(data), repr(sig))
	result = gpg3.GPG_verify2(data.encode('utf-8', 'surrogateescape'), sig.replace(boundary,'').encode('utf-8', 'surrogateescape'))
	if result == None:
		print("Warning: GPG_verify result is invalid", repr(result))
		return None
	#end if
	#(success, fpr, created, name) = result
	r = result
	if not hasattr(r, 'signatures') or len(r.signatures) == 0:
		success = 0
	else:
		success = r.signatures[0].summary & gpg3.gpg.constants.sigsum.VALID
		fpr = r.signatures[0].fpr
		created = r.signatures[0].timestamp
		if r.signatures[0].key:
			name = str(r.signatures[0].key)
		else:
			keys = gpg3.getKeylist(fpr)
			if len(keys) == 1:
				name = ', '.join(['{0} <{1}>'.format(uid.name, uid.email) for uid in keys[0].uids])
			elif len(keys) > 1:
				name = 'Multiple keys: '
				for key_i in keys:
					name += ', '.join(['{0} <{1}>'.format(uid.name, uid.email) for uid in key_i.uids]) + '; '
				#next key_i
			else:
				name = 'Unknown ' + fpr
			#end if
		#end if
	#end if
	if not success:
		print("Warning: Bad signature", repr(r), r.signatures[0].summary)
		#print('sigs', r.signatures, r.signatures[0].summary)
		return None
	#end if
	return (success, fpr, created, name, data)
#end def checkSig(plaintext)

def verifyClear(clearsigned):
	"""
	TODO: Check return
	"""
	if gpg3 == None:
		print("Warning: GPG3 module is missing and there's data to verify.")
		return None
	#end if
	return gpg3.GPG_verify(clearsigned)
#end def verifyClear(clearsigned)

def signDetatched(body, keyid):
	"""
	Outputs a PGP detached signature.
	
	"""
	if gpg3 == None:
		print("Warning: GPG3 module is missing and there's data to sign.")
		return None
	#end if
	return gpg3.GPG_sign(body, keyid)
#end def signDetatched(body, keyid)

def signOpaque(body, keyid):
	"""
	Outputs a PGP message with signature.
	
	"""
	if gpg3 == None:
		print("Warning: GPG3 module is missing and there's data to sign.")
		return None
	#end if
	return gpg3.GPG_sign(body, keyid, False)
#end def signOpaque(body, keyid)

def signTraditional(body, keyid):
	"""
	Outputs a traditional PGP clearsigned document.
	
	"""
	if gpg3 == None:
		print("Warning: GPG3 module is missing and there's data to sign.")
		return None
	#end if
	return gpg3.GPG_clearsign(body, keyid)
#end def signTraditional(body, keyid)

def signEncryptTraditional(body, to, keyid):
	"""
	Outputs a traditional PGP encrypted and signed document.
	
	"""
	if gpg3 == None:
		print("Warning: GPG3 module is missing and there's data to encrypt.")
		return None
	#end if
	return gpg3.GPG_sign_encrypt(body, to, keyid)
#end def signEncryptTraditional(body, to, keyid)

def encryptTraditional(body, to):
	"""
	Outputs a traditional PGP encrypted unsigned document.
	
	"""
	if gpg3 == None:
		print("Warning: GPG3 module is missing and there's data to encrypt.")
		return None
	#end if
	return gpg3.GPG_encrypt(body, to)
#end def signEncryptTraditional(body, to)

def importPubKey(key):
	"""
	Import a public key
	
	"""
	if gpg3 == None:
		print("Warning: GPG3 module is missing and there's data to import.")
		return None
	#end if
	return gpg3.GPG_import(key)
#end def importPubKey(key)

def mime_encode(m):
	"""
	FIXME: I think we need a better encoder probably.
	"""
	if isinstance(m, baseUnicode):
		m = MIMEText(m)
	#end if
	# May not be necessary?
	#email.encoders.encode_quopri(m)
	out = m.as_string()
	# Remove trailing \n
	#if out.endswith('\n'):
	#	out = out[:-1]
	#end if
	out = out.replace('\n', '\r\n')
	return out
#end def mime_encode(s)

def toCrLf(s):
	if isinstance(s, baseUnicode):
		if '\r\n' in s: return s
		return s.replace('\n', '\r\n')
	#end if
	if b'\r\n' in s: return s
	return s.replace(b'\n', b'\r\n')
#end def toCrLf(s)

def mime_encode_sign(m, sig, sigalg, retObject=False):
	"""
	FIXME: I think we need a better encoder probably.
	"""
	if isinstance(m, baseUnicode):
		m = MIMEText(m)
	#end if
	sigp = MIMEApplication(sig, 'pgp-signature', _encoder=email.encoders.encode_noop, name="signature.asc")
	sigp.add_header('Content-Disposition', 'attachment', filename='signature.asc')
	container = MIMEMultipart('signed', micalg='pgp-' + sigalg, protocol='application/pgp-signature', _subparts=[m, sigp])
	if retObject: return container
	return container.as_string()
#end def mime_encode_sign(s, sig, sigalg, retObject=False)

def mime_encode_encrypted(ciphertext):
	"""
	Unlike the other two, we return the object instead of the string.
	multipart/encrypted; protocol="application/pgp-encrypted"
	"""
	encrypted_msg = MIMEApplication(ciphertext, 'octet-stream', _encoder=email.encoders.encode_noop, name='encrypted.asc')
	encrypted_msg.add_header('Content-Disposition', 'inline', filename='encrypted.asc')
	version = MIMEApplication('Version: 1', 'pgp-encrypted', _encoder=email.encoders.encode_noop)
	container = MIMEMultipart('encrypted; protocol="application/pgp-encrypted"', _subparts=[version, encrypted_msg])
	return container
#end def mime_encode_encrypted(ciphertext)

sigalgs = {1:'md5', 2:'sha1', 3:'ripemd160', 8:'sha256', 9:'sha384', 10:'sha512', 11:'sha224'}

def replyQuote(body_data):
	"""
	Solve reply quoting.
	Input:
	test
	> asdf
	>> zxcv
	Output:
	> test
	>> asdf
	>>> zxcv
	"""
	#body_data = '>' + ('\n>'.join(body_data.split('\n')))
	lines = body_data.split('\n')
	output = ''
	for line in lines:
		if line[:1] == '>':
			output += '>'
		else:
			output += '> '
		#end if
		output += line + '\n'
	#next line
	# Strip the newline at the end
	if len(output) >= 1:
		output = output[:-1]
	#end if
	return output
#end def replyQuote(body_data)

def AttachBody(body, attach):
	"""
	Create a multipart mime plaintext body with attachments.
	body must be a string.
	attach must be an array of FileAttachment objects.
	Prints debug information.
	"""
	msg = MIMEMultipart()
	textPart = MIMEText(body)
	msg.attach(textPart)
	for attach_i in attach:
		ctype = attach_i.ctype
		maintype, subtype = 'application', 'octet-stream'
		if ctype and '/' in ctype:
			maintype, subtype = ctype.split('/', 1)
		#end if
		print('main:' + maintype + ', sub:' + subtype)
		if maintype == 'image':
			attachPart = MIMEImage(attach_i.value, subtype)
		elif maintype == 'audio':
			attachPart = MIMEAudio(attach_i.value, subtype)
		elif maintype == 'text':
			attachPart = MIMEText(attach_i.value.decode('us-ascii', 'surrogateescape'), subtype)
		elif maintype == 'application':
			attachPart = None
			if subtype == 'octet-stream' or subtype == 'x-unknown':
				# If they don't know what it is, it may be ascii text/plain.
				# But don't try to put binary or utf-8 into a text/plain
				try:
					attachPart = MIMEText(attach_i.value.decode('ascii', 'surrogateescape'), 'plain')
				except UnicodeEncodeError:
					attachPart = None
				#end try
			#end if
			if attachPart == None:
				encoder = email.encoders.encode_base64
				if isinstance(attach_i.value, baseUnicode):
					encoder = email.encoders.encode_7or8bit
				#end if
				attachPart = MIMEApplication(attach_i.value, subtype, encoder)
			#end if
		else:
			attachPart = MIMEBase(maintype, subtype)
			attachPart.set_payload(attach_i.value)
		#end if
		#attachPart.set_payload(attach_i.value)
		#encoders.encode_base64(attachPart)
		if attach_i.filename:
			attachPart.add_header('Content-Disposition', 'attachment', filename=attach_i.filename)
		#end if
		msg.attach(attachPart)
	#next attach_i
	return msg
#end def AttachBody(body, attach)

def prepareMessage(sfrom, to, cc, bcc, subject, body, attach=None, headers=None, crypto=False):
	"""
	Turn a set of inputs into a message
	"""
	to, sto = canonTo(to)
	cc, scc = canonTo(cc)
	#bcc, sbcc = canonTo(bcc)

	#msg = "From: %s\r\nTo: %s\r\nSubject: %s\r\n\r\n" % (sfrom, sto, subject)
	#msg += body
	# Attach is now an array of attachments.
	if attach and (attach[0] == None or len(attach[0].value) == 0):
		attach = None
	#end if
	if attach == None or len(attach) == 0:
		if isinstance(body, email.message.Message):
			msg = body
		else:
			msg = MIMEText(body)
		#end if
	else:
		# Use multipart to sort this out.
		msg = AttachBody(body, attach)
	#end if
	if crypto == False:
		msg['From'] = sfrom
		msg['To'] = sto
		if scc:
			msg['Cc'] = scc
			to.extend([recipient.strip() for recipient in scc.split(',')])
		#end if
		msg['Subject'] = subject
		if headers:
			xh = headers.split('\n')
			for line in xh:
				# FIXME: silent refusal of header
				if ':' not in line: continue
				# Only want left, right no exceptions
				k, v = line.split(':', 1)
				# FIXME: This could result in a multi-header situation.
				msg[k.strip()] = v.strip()
			#next line
		else:
			headers = ''
		#end if
		#debug += repr(to)
		#debug += repr(attach)

		date = email.utils.formatdate(time.time())
		msg['Date'] = date
	#end if
	return msg
#end def prepareMessage(sfrom, to, cc, bcc, subject, body, attach=None, headers=None)

def send(sfrom, to, cc, bcc, subject, body, attach=None, headers=None, server=None, maildir=None, fingerprint=None):
	"""
	Send a message.
	Currently it is hardcoded to send to localhost, HELO mail.altsci.com
	It requests TLS, but doesn't validate that it occurred or is valid cert.
	It sends the message plaintext body with optional attachments and headers.
	all values must be strings, not bytes.
	attach is an array or None.
	Attachments can be created using createAttachment(filename, 'main/subtype') 
	or FileAttachment(filename, data, contentType)
	headers is either None or a string. Don't use an array.
	It writes the message as well as it can to 
	~/Mail/sent-mail/cur/unixtime.pid.rand:2,S
	If this directory doesn't exist, it will crash before it sends, so don't 
	worry about double sending, send it again after you've created the 
	directory.
	Instead of exceptions (for the most part), it will return errors as strings.
	It returns "Success " + debug on success.
	"""
	import smtplib
	import random

	server_dest = ('localhost', 25, 'clientffff.bikeim.com')
	if server:
		hostname = socket.gethostname() 
		if (not hostname) or (hostname == 'localhost' and server[0] != 'localhost'):
			# If a user is talking to an external server and their hostname is localhost, they get to be clientxxxx.bikeim.com
			hostname = ('client{0:0x}.bikeim.com'.format(random.randint(0, 0xfffe)))
		#end if
		server_dest = (server[0], server[1], hostname)
	#end if
	debug = ''
	a = smtplib.SMTP(*server_dest)
	# FIXME: SSL certificate validation!
	try:
		a.starttls()
	except smtplib.SMTPServerDisconnected as ssd:
		return "Server disconnected. Do you have a good SMTP server?"
	#end try
	bincert = a.sock.getpeercert(True)
	server_fingerprint = hashlib.sha256(bincert).hexdigest()
	if fingerprint and (toUnicode(fingerprint).lower() != toUnicode(server_fingerprint)):
		a.close()
		# Cert pinning failed.
		return "Server certificate pinning failed. Server SHA256 fingerprint: {0}".format(server_fingerprint)
	#end if
	#a.set_debuglevel(1)

	# create a mime object from inputs
	msg = prepareMessage(sfrom, to, cc, bcc, subject, body, attach, headers)

	if maildir == None:
		path = os.path.join(os.path.expanduser('~'), 'Mail', 'sent-mail', 'new')
	else:
		path = os.path.join(maildir, 'sent-mail', 'new')
	#end if
	if not os.path.exists(path):
		try:
			os.makedirs(path)
		except (NotADirectoryError, FileExistsError):
			path = os.path.join(maildir, 'sent-mail2', 'new')
			if not os.path.exists(path):
				os.makedirs(path)
			#end if
		#end try
	#end if
	filename =  '{0:d}.{1:d}.{2}:2,S'.format(int(time.time()), os.getpid(), hex(random.randint(0, 1000000))[2:])

	# TODO: We have to be certain that body is always string, not bytes!
	open(os.path.join(path, filename), 'w').write(msg.as_string())
	
	to, sto = canonTo(to)
	cc, scc = canonTo(cc)
	bcc, sbcc = canonTo(bcc)
	recipients = to
	if cc:
		recipients += cc
	if bcc:
		recipients += bcc
	#return 'nope ' + sanitize(debug)
	try:
		a.sendmail(sfrom, recipients, msg.as_string())
	except smtplib.SMTPRecipientsRefused as e:
		return 'Recipients refused: ' + sanitize(str(e))
	except smtplib.SMTPHeloError as e:
		return 'SMTP HELO error: ' + sanitize(str(e))
	#end try
	a.quit()
	#header("Content-type:text/plain");
	return 'Success ' + sanitize(debug)
#end def send(sfrom, to, cc, bcc, subject, body, attach=None, headers=None)

class FileAttachment:
	"""
	A simple file attachment is made up of a string filename, a byte array 
	of data, and an optional content type.
	Equivalent to
	collections.namedtuple('FileAttachment', ['filename','value','ctype'])
	Classes support all versions of Python, namedtuple only supports 2.6+
	So I'm going to stick with this class.
	"""
	def __init__(self, filename, data, ctype=None):
		self.filename = filename
		self.value = data
		self.ctype = ctype
	#end def __init__(filename, data, ctype=None)
#end class FileAttachment

def createAttachment(filename, ctype=None):
	"""
	send takes an array of type FileAttachment which has data:
	filename
	value
	type
	value contains a binary.
	type is the content type like text/plain
	filename is optional to the FileAttachment, but not to this function.
	This function requires a valid filename.
	"""
	#TODO: Error handling
	output = None
	# TODO: Test binary and text attachments correctly...
	data = open(filename, 'rb').read()
	# TODO: I think this / stays because...
	output = FileAttachment(filename.split(os.sep)[-1], data, ctype)
	return output
#end def createAttachment(filename, ctype=None)

# Technically only 0x7f - 0xa0 are unique, but I found a \xae and the misuse of 
# ascii pretty much leaves us no other choice of guessing.
win1252_RE = re.compile(b'[' + b'\x7f\x80\x82\x83\x84\x85\x86\x87\x88\x89\x8a\x8b\x8c\x8e\x91\x92\x93\x94\x95\x96\x97\x98\x99\x9a\x9b\x9c\x9e\x9f\xa0\xa1\xa2\xa3\xa4\xa5\xa6\xa7\xa8\xa9\xaa\xab\xac\xad\xae' + bytes(range(0xf4, 0x100)) + b']')

def detectWindows1252(value):
	"""
	A common codec used by spammers and iRobot quite often in miscoded e-mail.
	We detect by finding characters between 0x7f and 0x9f inclusive.
	According to Wikipedia:
	Most modern web browsers and e-mail clients treat the MIME charset 
	ISO-8859-1 as Windows-1252 to accommodate such mislabeling. This is now 
	standard behavior in the HTML 5 specification, which requires that 
	documents advertised as ISO-8859-1 actually be parsed with the 
	Windows-1252 encoding.
		"""
	return bool(win1252_RE.search(value))
#end def detectWindows1252(value)

# Spam IP in Received.
ip_re = re.compile('\\[([^\\]]*)\\]')
def getReceivedIp(msg, server_name):
	"""
	Find the right ip address of the sender.
	Depends on the name of the receiving server.
	"""
	if isinstance(msg, dict): return None
	by_server_name = 'by {0}'.format(server_name)
	mip = None
	received_list = msg.get_all('received')
	if received_list == None: received_list = []
	for received in received_list:
		received = str(received)
		if by_server_name in received:
			x = ip_re.search(received)
			if x:
				mip = x.groups()[0]
				break
			#end if
		#end if
	#next received
	return mip
#end def getReceivedIp(msg, server_name)

# This function replaces message.walk
def walk(self):
	"""Walk over the message tree, yielding each subpart.

	The walk is performed in depth-first order.  This method is a
	generator.
	FIXME: empty file AssertionError
	"""
	x = self
	todo = []
	while x:
		yield x
		if x.is_multipart():
			y = x.get_payload()[:]
			x = y.pop(0)
			todo = y + todo
		else:
			if len(todo) > 0:
				x = todo.pop(0)
			else:
				break
			#end if
		#end if
	#loop
#end def walk(self)

def as_bytes(msg):
	"""
	msg.as_bytes() for multiple python versions
	"""
	if sys.hexversion >= 0x3040000:
		return msg.as_bytes()
	elif sys.hexversion >= 0x3000000:
		return msg.as_string().encode('ascii', 'surrogateescape')
	return msg.as_string()
#end def as_bytes(msg)

def main():
	filterAll = False
	filterFilename = 'list_filter1.json'
	if '-a' in sys.argv:
		filterAll = True
	#end if
	if '-f' in sys.argv:
		fpos = sys.argv.index('-f')
		if len(sys.argv) <= fpos + 1:
			print("Error: mail1.py -f filename filename is required.")
			return
		#end if
		filterFilename = sys.argv[fpos + 1]
		filterAll = True
	#end if
	Maildir = os.path.join(os.path.expanduser('~'), 'Mail')
	dotMaildir = os.path.join(os.path.expanduser('~'), '.maildir')
	config = getConfig()
	if 'Mail1' not in config:
		config['Mail1'] = {'inbox': 'inbox'}
	#end if
	currentDir = config['Mail1']['inbox']
	inboxCur = os.path.join(Maildir, currentDir, 'cur')
	# Command line invocation copies and filters.
	#moved2 = moveAll('/home/jvoss/.maildir/cur/', '/home/jvoss/Mail/badkm/cur/')
	#moved = moveAll('/home/jvoss/.maildir/new/', '/home/jvoss/Mail/badkm/cur/')
	#moved3 = moveAll('/home/jvoss/Mail/sent-mail/new/', '/home/jvoss/Mail/sent-mail/cur/')
	moved2 = moveAll(os.path.join(dotMaildir, 'cur'), inboxCur)
	moved = moveAll(os.path.join(dotMaildir, 'new'), inboxCur)
	moved3 = moveAll(os.path.join(Maildir, 'sent-mail', 'new', ''), 
				  os.path.join(Maildir, 'sent-mail', 'cur', ''))
	#print moved
	if filterAll:
		x, moved = filterDir(filterFilename, Maildir, currentDir)
		print(x)
	else:
		x, moved = filterList(filterFilename, moved, Maildir, currentDir)
		if x: print(x)
	#end if
	
#end def main()

if __name__ == '__main__':
	main()
#end if
